"""This software was initially develoepd by the US Department of the Interior
in 2014. It was modified and updated by a member of the oringal development team with 
support from CPC, Inc, in 2022. This software is provide 'as is'. 
The US Department of the Interior and CPC, Inc. give no warranty, expressed or implied, 
as to the accuracy, reliability,or completeness of this software. 
Although this software has been processed successfully on a computer system at CPC, Inc, no warranty
expressed or implied is made regarding the functioning of the software on
another system or for general or scientific purposes, nor shall the act of
distribution constitute any such warranty. This disclaimer applies both to
individual use of the software and aggregate use with other software.
"""
import sys
import os
import arcpy 
import numpy 
import datetime
import math
import pandas as pd
import glob
import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt
from arcpy.sa import *

class GammaReadings(object):
    def __init__(self):
        self.surveyLineData = {}

    def getLineNumbers(self, data):
        """Returns a set of line numbers from the data dictionary
        Arguments: 
        data - a list of dictionaries
        """
        lineNumberList = []
        for record in data:
            lineNumberList.append(record['lineNumber'])
        lineNumberSet = set(lineNumberList)
        return lineNumberSet

    def createSurveyLineData(self, data):
        """Splits the data dictionary up by line number and populates
        the surveyLineData dictionary.
        Arguments: 
        data - a list of dictionaries
        """
        lineNumbers = self.getLineNumbers(data)
        self.surveyLineData = {}
        for lineNumber in lineNumbers:
            self.surveyLineData['{}'.format(lineNumber)] = [record for record in data if record['lineNumber'] == lineNumber]		

    def removeOutlyingValuesSD(self, key):
        """Removes records that have a value outside two standard deviations of
        the mean for any key in the surveyLineData dictionary.  All records in all survey
        lines are used in the analysis.
        Arguments:
        data - list of dictionaries - list of dictionaries
        key - sting - key in the data dictionary to filter
        """
        for surveyLine in self.surveyLineData:
            numpyValues = numpy.array([float(record[key]) for record in self.surveyLineData[surveyLine]])
            standardDeviation = numpy.std(numpyValues)
            mean = numpy.mean(numpyValues)
            maxValue = mean + ( 2.0 * standardDeviation)
            minValue = mean - ( 2.0 * standardDeviation)
            self.surveyLineData[surveyLine] = [record for record in self.surveyLineData[surveyLine] if (float(record[key]) >= minValue and float(record[key]) <= maxValue)]

    def removeValueChangesGreaterThanNeighborhoodMean(self, key, change, back, forward):
        """Removes records that have a change in a key value greater than
        the specified allowable change from the mean of a neighborhood of points.
        Arguments:
        key - string - key in the surveyLineData dictionary to filter
        change - float - the maximum allowable change from record to record
        back - int - number previous records to include in the neighborhood mean value
        forward - int - number of following records to include in the negighborhood mean value
        """        
        for surveyLine in self.surveyLineData:
            for recordIndex in range(len(self.surveyLineData[surveyLine])):
                values = []
                neighborhood = range(recordIndex - back, recordIndex + forward + 1)
                neighborhood = [index for index in neighborhood if index != recordIndex]
                for searchIndex in neighborhood:
                    if searchIndex >= 0 and searchIndex < len(self.surveyLineData[surveyLine]):
                        values.append(self.surveyLineData[surveyLine][searchIndex][key])
                
                if len(values) > 0:
                    numpyValues = numpy.array([float(record) for record in values])
                    meanValue = numpyValues.mean()
                    deltaValue = self.surveyLineData[surveyLine][recordIndex][key] - meanValue
                    if abs(deltaValue) > change:
                        self.surveyLineData[surveyLine][recordIndex]['filter'] = True
                    else:
                        self.surveyLineData[surveyLine][recordIndex]['filter'] = False
                else:
                    self.surveyLineData[surveyLine][recordIndex]['filter'] = False                
            self.surveyLineData[surveyLine] = [record for record in self.surveyLineData[surveyLine] if record['filter'] == False]

    def calculateGradient(self, sensorNoise, back, forward):
        """Calculates the change in total field as the difference between the current and
        the mean of the surrounding total field readings. If the change in
        total field are less than the sensor noise it is reassigned the value of 0 and the
        change in total is assumed to be the result of noise. Does not maintain neighborhood
        size at the survey line ends.
        Aruguments: 
        sensorNoise - float - sensor noise in gamma
        back - int - number previous records to include in the mean background gamma
        forward - int - number of following records to include in the mean background gamma
        """		
        for surveyLine in self.surveyLineData:
            for recordIndex in range(len(self.surveyLineData[surveyLine])):
                values = []
                neighborhood = range(recordIndex - back, recordIndex + forward + 1)
                neighborhood = [index for index in neighborhood if index != recordIndex]
                for searchIndex in neighborhood:
                    if searchIndex >= 0 and searchIndex < len(self.surveyLineData[surveyLine]):
                        values.append(self.surveyLineData[surveyLine][searchIndex]['totalField'])
                if len(values) > 0:
                    numpyValues = numpy.array([float(record) for record in values])
                    meanValue = numpyValues.mean()
                    Gradient = self.surveyLineData[surveyLine][recordIndex]['totalField'] - meanValue
                    if abs(Gradient) < sensorNoise:
                        Gradient = 0
                    self.surveyLineData[surveyLine][recordIndex]['Gradient'] = Gradient
                else:
                    self.surveyLineData[surveyLine][recordIndex]['Gradient'] = 0

    def calculateSurveyLineGradients(self):
        
        i = 0
        for surveyLine in self.surveyLineData:
            i += len(self.surveyLineData[surveyLine]) - 1

        numpyGradientData = numpy.zeros(i, dtype={'names':['easting', 'northing', 'Gradient', 'lineNumber'], 'formats':['f8', 'f8', 'f8', 'a32']})

        index = 0
        for surveyLine in self.surveyLineData:
            multiplier = 0.0
            survey_line_len = len(self.surveyLineData[surveyLine])
            index_start = index
            index_stop = 0
            for recordIndex in range(1, survey_line_len):
                item2 = self.surveyLineData[surveyLine][recordIndex]
                item1 = self.surveyLineData[surveyLine][recordIndex - 1]
                difference = item2['totalField'] - item1['totalField']
                distance = math.sqrt(math.pow(item2['easting'] - item1['easting'], 2) + math.pow(item2['northing'] - item1['northing'], 2))

                numpyGradientData[index]['easting'] = (item2['easting'] + item1['easting']) / 2.0
                numpyGradientData[index]['northing'] = (item2['northing'] + item1['northing']) / 2.0
                numpyGradientData[index]['Gradient'] = difference / distance
                numpyGradientData[index]['lineNumber'] = item2['lineNumber']

                index += 1
                index_stop = index
                multiplier += distance

            multiplier = multiplier / (survey_line_len - 1)

            #arcpy.AddMessage('surveyLine: ' + str(surveyLine) + ', length: ' + str(survey_line_len) + ', multiplier: ' + str(multiplier))
            for recordIndex in range(index_start, survey_line_len - 1):
                numpyGradientData[recordIndex]['Gradient'] = numpyGradientData[recordIndex]['Gradient'] * multiplier
                
        return numpyGradientData

    def createNumpyArray(self, includeGradient):
        """Creates a numpy structured array from the surveyLineData dictionary
        Arguments: 
        includeGradient - bool - Set to true to include Gradient key
        Returns: 
        numpy structured array
        """
        data=[]				
        for surveyLine in self.surveyLineData:
            for record in self.surveyLineData[surveyLine]:
                data.append(record)
        if includeGradient:
            numpyData = numpy.zeros(len(data), dtype={'names':['easting', 'northing', 'totalField', 'lineNumber', 'altitude', 'Gradient'], 'formats':['f8', 'f8', 'f8', 'a32', 'i4', 'i4']})
            for index in range(len(data)):
                numpyData[index]['easting'] = data[index]['easting']
                numpyData[index]['northing'] = data[index]['northing']
                numpyData[index]['totalField'] = data[index]['totalField']
                numpyData[index]['lineNumber'] = data[index]['lineNumber']
##                numpyData[index]['date'] = data[index]['date']
##                numpyData[index]['time'] = data[index]['time']
                numpyData[index]['altitude'] = int(round(data[index]['altitude']))
                numpyData[index]['Gradient'] = int(round(data[index]['Gradient']))

        else:
            numpyData = numpy.zeros(len(data), dtype={'names':['easting', 'northing', 'totalField', 'lineNumber', 'date', 'time', 'altitude'], 'formats':['f8', 'f8', 'f8', 'a32', 'U10', 'U12', 'i4']})
            for index in range(len(data)):
                numpyData[index]['easting'] = data[index]['easting']
                numpyData[index]['northing'] = data[index]['northing']
                numpyData[index]['totalField'] = data[index]['totalField']
                numpyData[index]['lineNumber'] = data[index]['lineNumber']
                numpyData[index]['date'] = data[index]['date']
                numpyData[index]['time'] = data[index]['time']
                numpyData[index]['altitude'] = int(round(data[index]['altitude']))			

        return numpyData


    def readDataTableWithAltitude(self, inputTable, eastingField,
                                  northingField, totalFieldField, lineNumberField, dateField, timeField, altitudeField,
                                  altitudeUnit):
        """Reads a table with an arcpy search cursor and populates a dictionary
        for recorded magnetometer datasets using altitude.
        Arguments: 
        inputTable - string - path to a table
        eastingField - string - easting field name
        northingField - string - northing field name
        totalFieldField - string - recorded total field name
        lineNumberField - string - line number field name
        altitudeField - string - altitude field name
        altitudeUnit - string - altitude linear unit, either meter or feet
        """
        data = []
        unitConversions = {'meter':100, 'feet':30.48}
        with arcpy.da.SearchCursor(inputTable, [eastingField,
                                                northingField,
                                                totalFieldField,
                                                lineNumberField,
                                                dateField,
                                                timeField,
                                                altitudeField]) as cursor:
            for row in cursor:
                data.append({'easting':row[0],
                             'northing':row[1],
                             'totalField':row[2],
                             'lineNumber':row[3],
                             'date':row[4],
                             'time':row[5],                    
                             'altitude':row[6] * unitConversions[altitudeUnit]})
        self.createSurveyLineData(data)

    def readDataTableWithOutAltitude(self, inputTable, eastingField,
                                     northingField, totalFieldField, lineNumberField, dateField, timeField, sensorAltitude):
        """Reads a table with an arcpy search cursor and populates a dictionary
        for recorded magnetometer datasets without using altitude.
        Arguments: 
        inputTable - string - path to a table
        eastingField - string - easting field name
        northingField - string - northing field name
        totalFieldField - string - recorded gamma field name
        lineNumberField - string - line number field name
        sensorAltitude - double - constant sensor altitude in centimeters
        """
        data = []
        unitConversions = {'meter':100, 'feet':30.48}
        with arcpy.da.SearchCursor(inputTable, [eastingField,
                                                northingField,
                                                totalFieldField,
                                                lineNumberField, 
                                                dateField, 
                                                timeField]) as cursor:
            for row in cursor:
                data.append({'easting':row[0],
                             'northing':row[1],
                             'totalField':row[2],
                             'lineNumber':row[3],
                             'date':row[4],
                             'time':row[5],
                             'altitude':sensorAltitude})
        self.createSurveyLineData(data)

    def readSurveyLinesFromFeatureClass(self, inputFeatureClass):

        data = []
        with arcpy.da.SearchCursor(inputFeatureClass, ['SHAPE@X', 'SHAPE@Y', 'totalField', 'lineNumber', 'altitude']) as cursor:
            for row in cursor:
                data.append({'easting':row[0], 
                             'northing':row[1], 
                             'totalField':row[2],
                             'lineNumber':row[3],
                             'altitude':row[4]})
        self.createSurveyLineData(data)

class Toolbox(object): 
    def __init__(self): 
        """Define the toolbox (the name of the toolbox is the name of the .pyt file)."""
        self.label = 'Magnetometer Survey Toolbox'
        self.alias = '' 

        # List of tool classes associated with this toolbox 
        self.tools = [InputTool, ConfidenceTool, VisualizationTool, SurveyAreaTool, DeleteIdenticalTool, SmoothRawNavigationTool, SmoothRawAltitudeTool] 

class InputTool(object):
    def __init__(self):
        self.label = 'Input Tool'    
        self.description = ''
        self.canRunInBackground = True

    def getParameterInfo(self):

        param0 = arcpy.Parameter(displayName = 'Data Table', 
                                 name = 'inputTable', 
                                 datatype = 'Table', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param1 = arcpy.Parameter(displayName = 'Easting Field', 
                                 name = 'inputEasting', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param2 = arcpy.Parameter(displayName = 'Northing Field', 
                                 name = 'inputNorthing', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param3 = arcpy.Parameter(displayName = 'Raw Total Field (nT or Gamma)', 
                                 name = 'inputGamma', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param4 =  arcpy.Parameter(displayName = 'Altitude values included in table?', 
                                  name = 'useAltitudeField', 
                                  datatype = 'GPBoolean', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 
        
        param5 = arcpy.Parameter(displayName = 'Altitude Field', 
                                 name = 'inputAltitudeField', 
                                 datatype = 'Field', 
                                 parameterType = 'Optional', 
                                 direction = 'Input')

        param6 = arcpy.Parameter(displayName = 'Altitude is in Feet (otherwise in meters)', 
                                 name = 'useFeetAsAltitude', 
                                 datatype = 'GPBoolean', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param7 = arcpy.Parameter(displayName = 'Enter estimated altitude of sensor', 
                                  name = 'inputSensorAltitude', 
                                  datatype = 'Double', 
                                  parameterType = 'Optional', 
                                  direction = 'Input') 

        param8 = arcpy.Parameter(displayName= 'Line Number Field', 
                                 name = 'inputLineNumber', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 
                            
        param9 = arcpy.Parameter(displayName= 'Date Field', 
                                 name = 'inputDate', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param10 = arcpy.Parameter(displayName= 'Time Field', 
                                 name = 'inputTime', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param11 = arcpy.Parameter(displayName = 'Spatial Reference', 
                                 name = 'inputSpatialReference', 
                                 datatype = 'Spatial Reference', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param12 = arcpy.Parameter(displayName = 'Workspace',
                                  name = 'inputWorkspace', 
                                  datatype = 'Workspace', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param13 = arcpy.Parameter(displayName = 'Append rawPoints to Previous?', 
                                 name = 'yesAppend', 
                                 datatype = 'GPBoolean', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        # Default values 
        param1.value = "Field1"
        param2.value = "Field2"
        param3.value = "Field3"
        param4.value = True
        param5.value = "Field4"
        param6.value = True
        param7.value = 3
        param8.value = "Field5"
        param9.value = "Field6"
        param10.value = "Field7"
        param11.value = "NAD 1983 UTM Zone 17N"
        param13.value = False

        param1.parameterDependencies = [param0.name] 
        param2.parameterDependencies = [param0.name] 
        param3.parameterDependencies = [param0.name] 
        param5.parameterDependencies = [param0.name] 
        param8.parameterDependencies = [param0.name]
        param9.parameterDependencies = [param0.name]
        param10.parameterDependencies = [param0.name]
        return [param0, param1, param2, param3, param4, param5, param6, param7, param8, param9, param10, param11, param12, param13] 

    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.SetProduct('arcinfo') == "NotLicensed": 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters):
        if parameters[4].value == True:
            parameters[5].enabled = True
            parameters[7].enabled = False

        if parameters[4].value == False:
            parameters[5].enabled = False
            parameters[7].enabled = True

        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages): 
        """The source code of the tool."""
        # Set input parameters 
        inputTable = parameters[0].valueAsText 
        arcpy.AddMessage('Table: {}'.format(inputTable))

        inputEastingField = parameters[1].valueAsText 
        arcpy.AddMessage('Easting: {}'.format(inputEastingField))

        inputNorthingField = parameters[2].valueAsText 
        arcpy.AddMessage('Northing: {}'.format(inputNorthingField)) 

        inputGammaField = parameters[3].valueAsText 
        arcpy.AddMessage('Gamma: {}'.format(inputGammaField)) 

        useAltitudeField = parameters[4].value 
        arcpy.AddMessage('Use altitude field? {}'.format(str(useAltitudeField))) 
        
        if useAltitudeField:
            inputAltitudeField = parameters[5].valueAsText 
            arcpy.AddMessage('Altitude Field: {}'.format(inputAltitudeField))
        
        useFeetAsAltitude = parameters[6].value 
        arcpy.AddMessage('Altitude is measured in feet? {}'.format(str(useFeetAsAltitude))) 
        
        if useAltitudeField == False:
            unitConversions = {'meter':100, 'feet':30.48}
            if useFeetAsAltitude:
                inputSensorAltitude = parameters[7].value * unitConversions['feet']
                arcpy.AddMessage('Input sensor altitude: {}'.format(str(inputSensorAltitude)))
            else:
                inputSensorAltitude = parameters[7].value * unitConversions['meter']
                arcpy.AddMessage('Input sensor altitude: {}'.format(str(inputSensorAltitude)))

        inputLineNumberField = parameters[8].valueAsText 
        arcpy.AddMessage('Line Number Field: {}'.format(inputLineNumberField)) 

        inputDateField = parameters[9].valueAsText 
        arcpy.AddMessage('Line Number Field: {}'.format(inputDateField))

        inputTimeField = parameters[10].valueAsText 
        arcpy.AddMessage('Line Number Field: {}'.format(inputTimeField))

        inputSpatialReference = parameters[11].value 
        arcpy.AddMessage('Spatial reference: {}'.format(inputSpatialReference.name)) 

        outputWorkspace = parameters[12].valueAsText

        ##if not outputWorkspace.endswith('.gdb'):
            ##outputWorkspace = outputWorkspace + '.gdb'
        ## workspaceparts = os.path.split(outputWorkspace)

        arcpy.AddMessage('Workspace: {}'.format(outputWorkspace)) 

        appendrawPoints = parameters[13].value 
        arcpy.AddMessage('Append rawPoints? {}'.format(str(appendrawPoints)))

        # Set output workspace and overwrite 

        arcpy.env.workspace = r'{}'.format(outputWorkspace)
        arcpy.env.overwriteOutput = True

        # Create gammaReadings object 
        gammaReadings = GammaReadings() 

        # Read the data from the table 
        if useAltitudeField: 
            if useFeetAsAltitude: 
                gammaReadings.readDataTableWithAltitude(inputTable, inputEastingField,
                                                        inputNorthingField, inputGammaField,
                                                        inputLineNumberField, inputDateField, inputTimeField, inputAltitudeField, 'feet')
            else: 
                gammaReadings.readDataTableWithAltitude(inputTable, inputEastingField,
                                                        inputNorthingField, inputGammaField,
                                                        inputLineNumberField, inputDateField, inputTimeField, inputAltitudeField, 'meter') 
        else:
            if useFeetAsAltitude:
                gammaReadings.readDataTableWithOutAltitude(inputTable,
                                                           inputEastingField, inputNorthingField, inputGammaField,
                                                           inputLineNumberField, inputDateField, inputTimeField, inputSensorAltitude)
            else:
                gammaReadings.readDataTableWithOutAltitude(inputTable,
                                                           inputEastingField, inputNorthingField, inputGammaField,
                                                           inputLineNumberField, inputDateField, inputTimeField, inputSensorAltitude)
        arcpy.AddMessage('Read the data table')

        # copy rawPoints features if it exists to proceed appending
        if appendrawPoints == True:
            if arcpy.Exists("rawPoints"):
                arcpy.Rename_management("rawPoints","rawPointsapp")
                arcpy.management.Delete("rawPoints")
        if appendrawPoints == False:
            if arcpy.Exists("rawPoints"): 
                arcpy.management.Delete("rawPoints")

        arcpy.AddMessage('Cleaned Workspace')

        # Create a feature class from the gamma readings 
        gammaNumpyArray = gammaReadings.createNumpyArray(False) 

        out_table = "{}\{}".format(outputWorkspace,"rawPoints")
        arcpy.da.NumPyArrayToFeatureClass(gammaNumpyArray, out_table, ['easting', 'northing'], inputSpatialReference) 
        arcpy.AddMessage('Created rawPoints') 

        #append points
        if arcpy.Exists("rawPointsapp"):
            arcpy.management.Append("rawPoints", "rawPointsapp", "TEST")
            arcpy.AddMessage('Appending...')
            arcpy.management.Delete("rawPoints")
            arcpy.management.CopyFeatures("rawPointsapp", "rawPoints")
            arcpy.management.Delete("rawPointsapp")
            arcpy.AddMessage('Appended rawPoints')

        arcpy.env.workspace = outputWorkspace
        # convert points to lines >> surveyLines
        arcpy.PointsToLine_management('rawPoints', 'surveyLines', 'lineNumber')
        arcpy.AddMessage('Created surveyLines') 


class ConfidenceTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class."""
        self.label = 'Confidence Model'
        self.description = '' 
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""

        param0 = arcpy.Parameter(displayName = 'Input Survey Points Feature Class', 
                                 name = 'inputSurveyPoints', 
                                 datatype = 'DEFeatureClass', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 
        
        param1 = arcpy.Parameter(displayName = 'Filter elevations by standard deviations? (removes elevations outside of 2 standard deviations from the mean)', 
                                  name = 'filterElevationsSD', 
                                  datatype = 'GPBoolean', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param2 =  arcpy.Parameter(displayName = 'Filter elevations by maximum change from the neighborhood mean?', 
                                  name = 'filterElevationsMaxChange', 
                                  datatype = 'GPBoolean', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param3 = arcpy.Parameter(displayName = 'Points back to include in neighborhood for maximum change elevation filter ', 
                                  name = 'inputElevationPointsBack', 
                                  datatype = 'GPLong', 
                                  parameterType = 'Required', 
                                  direction = 'Input')

        param4 = arcpy.Parameter(displayName = 'Points forward to include in neighborhood for maximum change elevation filter', 
                                  name = 'inputElevationPointsForward', 
                                  datatype = 'GPLong', 
                                  parameterType = 'Required', 
                                  direction = 'Input')

        param5 = arcpy.Parameter(displayName = 'Maximum elevation change from neighborhood average (centimeters)', 
                                 name = 'maximumChange', 
                                 datatype = 'Double', 
                                 parameterType = 'Optional', 
                                 direction = 'Input') 

        param6 = arcpy.Parameter(displayName = 'Survey Area', 
                                  name = 'inputSurveyArea', 
                                  datatype = 'Feature Class', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param7 = arcpy.Parameter(displayName = 'Magnetic Moment', 
                                  name = 'inputMagneticMoment', 
                                  datatype = 'Double', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param8 = arcpy.Parameter(displayName = 'Sensor Noise', 
                                  name = 'inputSensorNoise', 
                                  datatype = 'Double', 
                                  parameterType = 'Required', 
                                  direction = 'Input')

        param9 = arcpy.Parameter(displayName = 'Mass Threshold', 
                                  name = 'inputMassThreshold', 
                                  datatype = 'Double', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param10 = arcpy.Parameter(displayName = 'Points back to include in neighborhood for spatial gradient calculation', 
                                  name = 'inputGradientPointsBack', 
                                  datatype = 'GPLong', 
                                  parameterType = 'Required', 
                                  direction = 'Input')

        param11 = arcpy.Parameter(displayName = 'Points forward to include in neighborhood for spatial gradeint calculation', 
                                  name = 'inputGradientPointsForward', 
                                  datatype = 'GPLong', 
                                  parameterType = 'Required', 
                                  direction = 'Input')

        param12 = arcpy.Parameter(displayName = 'Cell size for output rasters', 
                                  name = 'inputCellSize', 
                                  datatype = 'GPLong', 
                                  parameterType = 'Required', 
                                  direction = 'Input')

        param13 =  arcpy.Parameter(displayName = 'Send Outputs to same geodatabase as input?', 
                                  name = 'useInputGeodatabase', 
                                  datatype = 'GPBoolean', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param14 = arcpy.Parameter(displayName = 'Existing Output File Geodatabase',
                                  name = 'outputWorkspace', 
                                  datatype = 'Workspace', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        # Default values 
        param1.value = False
        param2.value = False
        param3.value = 1
        param4.value = 1
        param5.value = 1
        param7.value = 50
        param8.value = 5
        param9.value = 100
        param10.value = 1
        param11.value = 1
        param12.value = 1
        param13.value = True

        # Dependencies 
        return [param0, param1, param2, param3, param4, param5, param6, param7, param8, 
                param9, param10, param11, param12, param13, param14] 


    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.CheckExtension('Spatial') != 'Available': 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        """Modify the values and properties of parameters before internal 
        validation is performed.  This method is called whenever a parameter 
        has been changed."""
        if parameters[2].value == True:
            parameters[3].enabled = True
            parameters[4].enabled = True
            parameters[5].enabled = True
            
        if parameters[2].value == False:
            parameters[3].enabled = False
            parameters[4].enabled = False
            parameters[5].enabled = False

        if (parameters[13].value == True):
            parameters[14].enabled = False

            if (parameters[0].altered and len(parameters[0].valueAsText) > 0) : 
                workspaceparts = os.path.split(parameters[0].valueAsText)
                parameters[14].value = workspaceparts[0]
            else :
                parameters[14].value = ''
        else :
            parameters[14].enabled = True

        return

    def updateMessages(self, parameters): 
        """Modify the messages created by internal validation for each tool 
        parameter.  This method is called after internal validation."""
        try:
            inputFeatureClass = parameters[0].valueAsText
            if len(inputFeatureClass) > 0: 
                workspaceparts = os.path.split(inputFeatureClass)
            
                arcpy.env.workspace = workspaceparts[0]
                desc = arcpy.Describe(inputFeatureClass)

                if (desc.shapeType != 'Point'):
                    parameters[0].setErrorMessage('Incorrect input shape type, should be points feature class')
                else :
                    fieldList = arcpy.ListFields(inputFeatureClass)
                    b_gamma = False
                    b_line = False

                    for field in fieldList:
                        if field.name == 'totalField':
                            b_gamma = True
                        elif field.name == 'lineNumber':
                            b_line = True

                    if b_gamma == False or b_line == False or b_alt == False:
                        parameters[0].setErrorMessage('Not all required input fields present lineNumber and totalField')

        except Exception as e:
            return
        return

    def execute(self, parameters, messages): 
        """The source code of the tool."""
        # Set input parameters 
        inputFeatureClass = parameters[0].valueAsText 
        arcpy.AddMessage('input Survey Features: {}'.format(inputFeatureClass))

        filterElevationsSD = parameters[1].value
        arcpy.AddMessage('Filter elevations based on Standard Deviations? {}'.format(str(filterElevationsSD)))

        filterElevationsMaxChange = parameters[2].value 
        arcpy.AddMessage('Filter elevations based on maximum change? {}'.format(str(filterElevationsMaxChange))) 

        if filterElevationsMaxChange:
            inputElevationPointsBack = int(parameters[3].value)
            arcpy.AddMessage('Elevation points back: {}'.format(str(inputElevationPointsBack)))
            
            inputElevationPointsForward = int(parameters[4].value)
            arcpy.AddMessage('Elevation points forward: {}'.format(str(inputElevationPointsForward)))
            
            maximumChange = parameters[5].value
            arcpy.AddMessage('Maximum elevation change from neighborhood mean: {}'.format(str(maximumChange)))
    
        inputSurveyArea = parameters[6].valueAsText 
        arcpy.AddMessage('Survey area: {}'.format(inputSurveyArea)) 

        inputMagneticMoment = parameters[7].value
        arcpy.AddMessage('Magnetic moment: {}'.format(str(inputMagneticMoment))) 

        inputSensorNoise = parameters[8].value
        arcpy.AddMessage('Sensor noise: {}'.format(str(inputSensorNoise))) 

        inputMassThreshold = parameters[9].value
        arcpy.AddMessage('Mass threshold: {}'.format(str(inputMassThreshold)))

        inputGradientPointsBack = int(parameters[10].value)
        arcpy.AddMessage('Points back for spatial gradient calculation: {}'.format(str(inputGradientPointsBack)))

        inputGradientPointsForward = int(parameters[11].value)
        arcpy.AddMessage('Points forward for spatial calculation: {}'.format(str(inputGradientPointsForward)))

        inputCellSize = parameters[12].value
        arcpy.AddMessage('Cell size for output rasters: {}'.format(str(inputCellSize)))

        outputWorkspace = parameters[14].valueAsText

        arcpy.AddMessage('outputWorkspace: {}'.format(outputWorkspace))

        # Set arcpy environment variables 
        arcpy.env.workspace = outputWorkspace 
        arcpy.env.overwriteOutput = True
        arcpy.env.mask = inputSurveyArea 
        arcpy.env.extent = inputSurveyArea

        arcpy.AddMessage('Set environment variables')         

        suffix = ''
        delta_suffix = ''
        delta_suffix_2 = ''

        if filterElevationsSD:
            delta_suffix = delta_suffix + '_FSD'

        if filterElevationsMaxChange:
            delta_suffix = delta_suffix + '_FMC_' + str(inputElevationPointsBack) + '_' + str(inputElevationPointsForward) + '_' + str(int(maximumChange))

        delta_suffix = delta_suffix + '_N_' + str(int(inputSensorNoise)) + '_DR_' + str(int(inputGradientPointsBack)) + '_' + str(int(inputGradientPointsForward)) 
        delta_suffix_2 = delta_suffix + '_CS_' + str(int(inputCellSize)) 

        suffix = delta_suffix_2 + '_MM_' + str(int(inputMagneticMoment)) + '_T_' + str(int(inputMassThreshold)) 
        
        # Create the logTable
        now = datetime.datetime.now()
        logTable = 'ConfidenceLog' + suffix
        arcpy.CreateTable_management(outputWorkspace, logTable)
        arcpy.AddField_management(logTable, 'Process', 'TEXT')
        arcpy.AddField_management(logTable, 'Completed', 'TEXT')

        # Create an insert cursor to write to the logTable
        logCursor = arcpy.da.InsertCursor(logTable,('Process', 'Completed'))

        # Write the input parameters to the logTable
        if filterElevationsSD:
            logCursor.insertRow(('Filter elevations based on Standard Deviations? {}'.format(str(filterElevationsSD)), datetime.datetime.now()))

        if filterElevationsMaxChange:
            logCursor.insertRow(('Filter elevations based on maximum change? {}'.format(str(filterElevationsMaxChange)), datetime.datetime.now()))
            logCursor.insertRow(('Elevation points back: {}'.format(str(inputElevationPointsBack)), datetime.datetime.now()))
            logCursor.insertRow(('Elevation points forward: {}'.format(str(inputElevationPointsForward)), datetime.datetime.now()))
            logCursor.insertRow(('Maximum elevation change from neighborhood mean: {}'.format(str(maximumChange)), datetime.datetime.now()))
            
        logCursor.insertRow(('Survey area: {}'.format(inputSurveyArea), datetime.datetime.now()))
        logCursor.insertRow(('Magnetic moment: {}'.format(str(inputMagneticMoment)), datetime.datetime.now()))
        logCursor.insertRow(('Sensor noise: {}'.format(str(inputSensorNoise)), datetime.datetime.now()))
        logCursor.insertRow(('Mass threshold: {}'.format(str(inputMassThreshold)), datetime.datetime.now()))
        logCursor.insertRow(('Points back for spatial gradient calculation: {}'.format(str(inputGradientPointsBack)), datetime.datetime.now()))
        logCursor.insertRow(('Points forward for spatial gradient calculation: {}'.format(str(inputGradientPointsForward)), datetime.datetime.now()))
        logCursor.insertRow(('Cell size for output rasters: {}'.format(str(inputCellSize)), datetime.datetime.now()))

        # Check out the spatial analyst extension 
        arcpy.CheckOutExtension('Spatial') 
        arcpy.AddMessage('Checked out Spatial Analyst extension') 
        logCursor.insertRow(('Checked out Spatial Analyst extension', datetime.datetime.now()))

        # Create gammaReadings object 
        gammaReadings = GammaReadings() 

        # Read the data from the table 
        gammaReadings.readSurveyLinesFromFeatureClass(inputFeatureClass)

        arcpy.AddMessage('Read the data table')
        logCursor.insertRow(('Read the data table', datetime.datetime.now()))

        # Filter elevation values outside of 2 standard deviations
        if filterElevationsSD:
            gammaReadings.removeOutlyingValuesSD('altitude')
            arcpy.AddMessage('Filtered altitude values based on standard deviations')
            logCursor.insertRow(('Filtered altitude values based on standard deviations', datetime.datetime.now()))

        # Filter elevation values greater than the neighborhood mean if selected 
        if filterElevationsMaxChange:
            gammaReadings.removeValueChangesGreaterThanNeighborhoodMean('altitude',
                                                                        maximumChange, inputElevationPointsBack,
                                                                        inputElevationPointsForward)
            arcpy.AddMessage('Filtered elevations based on maximum change') 
            logCursor.insertRow(('Filtered elevations based on maximum change', datetime.datetime.now()))

        # Calculate spatial gradient values 
        gammaReadings.calculateGradient(inputSensorNoise, inputGradientPointsBack, inputGradientPointsForward)
        arcpy.AddMessage('Calculated Spatial Gradient') 
        logCursor.insertRow(('Calculated Spatial Gradient', datetime.datetime.now()))

        inputSpatialReference = arcpy.Describe(inputFeatureClass).spatialReference

        # Create a feature class from the gamma readings 
        gammaNumpyArray = gammaReadings.createNumpyArray(True)
        Gradient =  'integerGradient' + delta_suffix
        outputFeatureClass = os.path.join(outputWorkspace, Gradient)
        if arcpy.Exists(outputFeatureClass):
            arcpy.management.Delete(outputFeatureClass)
        arcpy.da.NumPyArrayToFeatureClass(gammaNumpyArray, outputFeatureClass, ['easting', 'northing'], inputSpatialReference) 
        arcpy.AddMessage('Created Gradient Integer Output') 
        logCursor.insertRow(('Created Gradient Integer Output', datetime.datetime.now()))

        del gammaNumpyArray
        del gammaReadings

        # Allocate Gradient Values as a raster 
        arcpy.AddMessage('Set Output Extent')
        integerGradientSurface = EucAllocation(Gradient, '', '', inputCellSize, 'Gradient') 
##        spatialGradientSurface.save('integerGradient' + delta_suffix_2) 
        arcpy.AddMessage('Created Confidence Model Integer Gradient Surface')
        logCursor.insertRow(('Created Confidence Model Integer Gradient Surface', datetime.datetime.now()))

        # Create the absolute value of spatialGradient
        absintegerGradientSurface = Abs(integerGradientSurface) 
        absintegerGradientSurface.save('absintegerGradient' + delta_suffix_2) 
        arcpy.AddMessage('Created abs Confidence Model Integer Gradient Surface') 
        logCursor.insertRow(('Created abs Confidence Model Integer Gradient Surface', datetime.datetime.now()))

        # Set the geoprocessing snap raster to absintegerGradient 
        arcpy.env.snapRaster = 'absintegerGradient' + delta_suffix_2
        arcpy.AddMessage('Set snap raster') 
        logCursor.insertRow(('Set snap raster', datetime.datetime.now()))

        # Allocate altitude values as a raster 
        altitudeCentimeters = EucAllocation(Gradient, '', '', inputCellSize, 'altitude') 
##        altitudeCentimeters.save('altitudeCentimeters' + delta_suffix_2) 
        arcpy.AddMessage('Created altitude') 
        logCursor.insertRow(('Created altitude', datetime.datetime.now()))

        # Create a horizontal euclidean distance raster from the input source points 
        horizontalDistance = EucDistance(Gradient, '', inputCellSize) 

        horizontalCentimeters = Int(horizontalDistance * inputSpatialReference.metersPerUnit * 100) 
##        horizontalCentimeters.save('horizontalCentimeters' + delta_suffix_2) 
        arcpy.AddMessage('Created horizontalCentimeters') 
        logCursor.insertRow(('Created horizontalCentimeters', datetime.datetime.now()))

        # Calculate the final distance raster in centimeters 
        centimeters = Int((horizontalCentimeters**2 + altitudeCentimeters**2)**0.5) 
        centimeters.save('sensorDistance' + delta_suffix_2) 
        arcpy.AddMessage('Created centimeters with altitude') 
        logCursor.insertRow(('Created centimeters with altitude field values', datetime.datetime.now()))

        # Create mass (kg) of potentially missed objects 
        potentialMassMissed = Int((inputSensorNoise * 0.000001 * (centimeters)**3) / (inputMagneticMoment)) 
        potentialMassMissed.save('potentialMassMissed' + suffix) 
        arcpy.AddMessage('Created potentialMassMissed') 
        logCursor.insertRow(('Created potentialMassMissed', datetime.datetime.now()))

        try:
            arcpy.BuildRasterAttributeTable_management('potentialMassMissed' + suffix, "Overwrite")
        except:
            arcpy.AddMessage("Build Raster Attribute Table failed for potential mass missed.")

        # Create mass (kg) of potentially found objects 
        potentialMassObserved = Int((absintegerGradientSurface * 0.000001 * (centimeters)**3) / (inputMagneticMoment)) 
        potentialMassObserved.save('potentialMassObserved' + suffix) 
        arcpy.AddMessage('Created potentialMassObserved') 
        logCursor.insertRow(('Created potentialMassObserved', datetime.datetime.now()))

        try:
            arcpy.BuildRasterAttributeTable_management('potentialMassObserved' + suffix, "Overwrite")
        except:
            arcpy.AddMessage("Build Raster Attribute Table failed for potential mass observed.")


        # Reclassify potential mass missed raster
        if inputMassThreshold > potentialMassMissed.minimum:
            remapRange = RemapRange([[0,inputMassThreshold, 1], [inputMassThreshold, 9999999, 0]])
            confidenceResult = Reclassify(potentialMassMissed, 'Value', remapRange) 
            confidenceResult.save('confidenceResult' + suffix)
        else:
            remapRange = RemapRange([[inputMassThreshold, 9999999, 0]])
            confidenceResult = Reclassify(potentialMassMissed, 'Value', remapRange) 
            confidenceResult.save('confidenceResult' + suffix)           
        arcpy.AddMessage('Created Confidence Result')
        logCursor.insertRow(('Created Confidence Result', datetime.datetime.now()))

        # Display the percentage of the survey area adequatly surveyed
        fields = ['VALUE', 'COUNT']
        data = {r[0]:r[1] for r in arcpy.da.SearchCursor('confidenceResult' + suffix, fields)}
        keys = data.keys()
        if 0 in keys:
            if 1 in keys:
                totalCount = data[0] + data[1]
                percentCovered = (data[1] / totalCount)
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('Percent Adequatly Surveyed: \n{:.2%}'.format(percentCovered))
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('*****************')
                logCursor.insertRow(('Percent Adequatly Surveyed: {:.2%}'.format(percentCovered), datetime.datetime.now()))

            else:
                percentCovered = 0
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('Percent Adequatly Surveyed: \n{:.2%}'.format(percentCovered))
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('*****************')
                logCursor.insertRow(('Percent Adequatly Surveyed: {:.2%}'.format(percentCovered), datetime.datetime.now()))

        elif 0 not in keys:
            if 1 in keys:
                percentCovered = 1
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('Percent Adequatly Surveyed: \n{:.2%}'.format(percentCovered))
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('*****************')
                logCursor.insertRow(('Percent Adequatly Surveyed: {:.2%}'.format(percentCovered), datetime.datetime.now()))

        # Delete the logCursor object
        del logCursor
        arcpy.CheckInExtension('Spatial') 

class VisualizationTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class)"""
        self.label = 'Visualization'
        self.description = '' 
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""
        param0 = arcpy.Parameter(displayName = 'Input Survey Points Feature Class', 
                                 name = 'inputSurveyPoints', 
                                 datatype = 'DEFeatureClass', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param1 = arcpy.Parameter(displayName = 'Select a Surface creation Method',
                                 name = 'surfaceMethod',
                                 datatype = 'GPString',
                                 parameterType = 'Required',
                                 direction = 'Input')

        param2 = arcpy.Parameter(displayName = 'Survey Area', 
                                 name = 'inputSurveyArea', 
                                 datatype = 'Feature Class', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param3 = arcpy.Parameter(displayName = 'Min Curvature Weight', 
                                 name = 'splineWeight', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input')
        
        param4 = arcpy.Parameter(displayName = 'Min Curvature Points', 
                                 name = 'splinePoints', 
                                 datatype = 'Long', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 
        
        param5 = arcpy.Parameter(displayName = 'Total Field Contour Interval (nT or gamma)', 
                                 name = 'inputRawContourInterval', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param6 = arcpy.Parameter(displayName = 'Spatial Gradient Contour Interval (nT or gamma)', 
                                 name = 'inputGradientContourInterval', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param7 = arcpy.Parameter(displayName = 'Cell Size for output rasters',
                                 name = 'inputCellSize',
                                 datatype = 'GPLong',
                                 parameterType = 'Required',
                                 direction = 'Input')

        param8 =  arcpy.Parameter(displayName = 'Send Outputs to same geodatabase as input?', 
                                  name = 'useInputGeodatabase', 
                                  datatype = 'GPBoolean', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param9 = arcpy.Parameter(displayName = 'Existing Output File Geodatabase',
                                  name = 'outputWorkspace', 
                                  datatype = 'Workspace', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        # Default values 
        param3.value = 0.1
        param4.value = 12
        param5.value = 10.0
        param6.value = 5.0
        param7.value = 1
        param8.value = True

        param1.filter.type = "ValueList"

        if arcpy.CheckExtension('3D') != "Available":
            param1.filter.list = ["Minimum Curvature"]
            param1.value = "Minimum Curvature"
        else:
            param1.filter.list = ["Minimum Curvature", "TIN"]
            param1.value = "TIN"

        # Dependencies 
        return [param0, param1, param2, param3, param4, param5, param6, param7, param8, param9] 

    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.CheckExtension('Spatial') != "Available": 
                raise Exception
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        """Modify the values and properties of parameters before internal 
        validation is performed.  This method is called whenever a parameter 
        has been changed."""

        if (parameters[1].value == "Minimum Curvature") :
            parameters[3].enabled = True
            parameters[4].enabled = True
        else :
            parameters[3].enabled = False
            parameters[4].enabled = False

        if (parameters[5].value < 0) :
            parameters[5].value = abs(parameters[5].value)

        if (parameters[6].value < 0) :
            parameters[6].value = abs(parameters[6].value)

        if (parameters[8].value == True):
            parameters[9].enabled = False

            if (parameters[0].altered and len(parameters[0].valueAsText) > 0) : 
                workspaceparts = os.path.split(parameters[0].valueAsText)
                parameters[9].value = workspaceparts[0]
            else :
                parameters[9].value = ''
        else :
            parameters[9].enabled = True

        return

    def updateMessages(self, parameters): 
        """Modify the messages created by internal validation for each tool 
        parameter.  This method is called after internal validation."""
        try:
            inputFeatureClass = parameters[0].valueAsText
            if len(inputFeatureClass) > 0: 
                desc = arcpy.Describe(inputFeatureClass)

                if (desc.shapeType != 'Point'):
                    parameters[0].setErrorMessage('Incorrect input shape type, should be points feature class')
                else :
                    fieldList = arcpy.ListFields(inputFeatureClass)
                    b_gamma = False
                    b_line = False

                    for field in fieldList:
                        if field.name == 'totalField':
                            b_gamma = True
                        elif field.name == 'lineNumber':
                            b_line = True

                    if b_gamma == False or b_line == False or b_alt == False:
                        parameters[0].setErrorMessage('Not all required input fields present lineNumber and totalField')

        except Exception as e:
            return

        return

    def execute(self, parameters, messages): 
        """The source code of the tool."""
        # Set input variables 
        inputFeatureClass = parameters[0].valueAsText 
        arcpy.AddMessage('input Survey Features: {}'.format(inputFeatureClass))

        outputWorkspace = parameters[9].valueAsText

        inputSurfaceMethod = parameters[1].value
        arcpy.AddMessage('inputSurfaceMethod: ' + inputSurfaceMethod)

        inputSurveyArea = parameters[2].valueAsText 
        arcpy.AddMessage('inputSurveyArea: ' + inputSurveyArea)

        if (inputSurfaceMethod == "Minimum Curvature") :
            splineWeight = parameters[3].value
            splinePoints = parameters[4].value
            arcpy.AddMessage('Weight: ' + str(splineWeight))
            arcpy.AddMessage('Points: ' + str(splinePoints))

        inputRawContourInterval = parameters[5].value
        arcpy.AddMessage('Total Field Contour Interval: ' + str(inputRawContourInterval))

        inputGradientContourInterval = parameters[6].value
        arcpy.AddMessage('Spatial Gradient Contour Interval: ' + str(inputGradientContourInterval))

        inputCellSize = parameters[7].value
        arcpy.AddMessage('inputCellSize: ' + str(inputCellSize))

        # Set arcpy environment variables 
        arcpy.env.workspace = outputWorkspace 
        arcpy.env.overwriteOutput = True
        arcpy.env.extent = inputSurveyArea
        arcpy.AddMessage('Scratch Workspace Folder: ' + arcpy.env.scratchFolder)
        arcpy.AddMessage('Set environment variables') 

        suffix = ''
        suffix_w_raw_contour = ''
        suffix_w_gradient_contour = ''
        suffix_w_both_contours = ''

        if (inputSurfaceMethod == "Minimum Curvature") :
            suffix = '_MinCurv_PT_' + str(int(splinePoints)) + '_CS_' + str(int(inputCellSize))
        elif (inputSurfaceMethod == "TIN") :
            suffix = '_TIN_CS_' + str(int(inputCellSize))

        suffix_w_raw_contour = suffix + '_RC_' + str(int(inputRawContourInterval))
        suffix_w_gradient_contour = suffix + '_GC_' + str(int(inputGradientContourInterval))
        suffix_w_both_contour = suffix + '_RC_' + str(int(inputRawContourInterval)) + '_GC_' + str(int(inputGradientContourInterval))

        # Create the logTable
        logTable = 'VisualizationLog' + suffix_w_both_contour
        arcpy.CreateTable_management(outputWorkspace, logTable)
        arcpy.AddField_management(logTable, 'Process', 'TEXT')
        arcpy.AddField_management(logTable, 'Completed', 'TEXT')

        # Create an insert cursor to write to the logTable
        logCursor = arcpy.da.InsertCursor(logTable,('Process', 'Completed'))

        # Write the input parameters to the logTable
        logCursor.insertRow(('Survey Points: {}'.format(inputFeatureClass), datetime.datetime.now()))
        logCursor.insertRow(('Survey area: {}'.format(inputSurveyArea), datetime.datetime.now()))
        logCursor.insertRow(('Surface Generation Method: ' + inputSurfaceMethod, datetime.datetime.now()))

        if (inputSurfaceMethod == "Minimum Curvature") :
            logCursor.insertRow(('Spline Weight: {}'.format(str(splineWeight)), datetime.datetime.now()))
            logCursor.insertRow(('Spline Points: {}'.format(str(splinePoints)), datetime.datetime.now()))

        logCursor.insertRow(('Raw Gamma Contour interval: {}'.format(str(inputRawContourInterval)), datetime.datetime.now()))
        logCursor.insertRow(('Gradient Gamma Contour interval: {}'.format(str(inputGradientContourInterval)), datetime.datetime.now()))
        logCursor.insertRow(('Cell size for output rasters: {}'.format(str(inputCellSize)), datetime.datetime.now()))

        # Check out the spatial analyst extension 
        arcpy.CheckOutExtension('Spatial') 
        arcpy.AddMessage('Checked out Spatial Analyst extension')
        logCursor.insertRow(('Checked out Spatail Analyst extension', datetime.datetime.now()))

        # Create gammaReadings object 
        gammaReadings = GammaReadings() 

        gammaReadings.readSurveyLinesFromFeatureClass(inputFeatureClass)

        # Create a feature class from the total field readings 
        gammaNumpyArray = gammaReadings.calculateSurveyLineGradients() 
        gradientGammaFeatureClass = os.path.join(outputWorkspace, 'spatialGradient')
        if not arcpy.Exists(gradientGammaFeatureClass):
            arcpy.da.NumPyArrayToFeatureClass(gammaNumpyArray, gradientGammaFeatureClass, ['easting', 'northing'], arcpy.Describe(inputFeatureClass).spatialReference) 
            arcpy.AddMessage('Created spatialGradient Feature Class') 
        else :
            arcpy.AddMessage('spatialGradient Feature Class already exists') 
 

        rawGammaSurface = None
        gradientGammaSurface = None

        # Interpolate a surface of total field values
        if (inputSurfaceMethod == "Minimum Curvature") :
            # Set the geoprocessing mask 
            arcpy.env.mask = inputSurveyArea 
            arcpy.AddMessage('Set analysis mask')
            logCursor.insertRow(('Set analysis mask', datetime.datetime.now()))
            
            totalFieldSurface = 'totalField' + suffix
            totalField = Spline(inputFeatureClass, 'totalField', str(inputCellSize), 'REGULARIZED', splineWeight, splinePoints)
            totalField.save(totalFieldSurface)		
            arcpy.AddMessage('Interpolated total field by Minimun Curvature') 
            logCursor.insertRow(('Interpolated total field by Minimum Curvature', datetime.datetime.now()))

            spatialGradientSurface = 'spatialGradient' + suffix
            spatialGradient = Spline(gradientGammaFeatureClass, 'Gradient', str(inputCellSize), 'REGULARIZED', splineWeight, splinePoints)
            spatialGradient.save(spatialGradientSurface)		
            arcpy.AddMessage('Interpolated spatial gradient by Minimum Curvature') 
            logCursor.insertRow(('Interpolated spatial gradient by Minimum Curvature', datetime.datetime.now()))

        elif (inputSurfaceMethod == "TIN") :
            arcpy.CheckOutExtension("3D")
            arcpy.AddMessage('Checked out 3D extension') 
            logCursor.insertRow(('Checked out 3D extension', datetime.datetime.now()))

            try:
                desc = arcpy.Describe(inputFeatureClass)
                inputSpatialReference = desc.spatialReference
                rawTIN = arcpy.env.scratchFolder + '\\rawTin'
                arcpy.AddMessage('Creating TIN dataset: ' + rawTIN)
                arcpy.ddd.CreateTin(rawTIN, inputSpatialReference, inputFeatureClass + " totalField masspoints", "DELAUNAY") 
                arcpy.AddMessage('TIN dataset created')
                logCursor.insertRow(('TIN dataset created', datetime.datetime.now()))

                totalFieldSurface = 'totalField' + suffix
                totalFieldSurfaceTemp = totalFieldSurface + '_temp' 
                arcpy.env.mask = inputSurveyArea 
                arcpy.TinRaster_3d(rawTIN, totalFieldSurfaceTemp, "FLOAT", "LINEAR", "CELLSIZE " + str(inputCellSize), "1")
                
                rGS = ExtractByMask(totalFieldSurfaceTemp, inputSurveyArea)
                rGS.save(totalFieldSurface)
                
                arcpy.Delete_management(totalFieldSurfaceTemp)
                arcpy.AddMessage('TIN surface raster created')

                arcpy.Delete_management(rawTIN)
                arcpy.AddMessage('TIN deleted:' + rawTIN)

                gradientTIN = arcpy.env.scratchFolder + '\\gradientTin'
                arcpy.AddMessage('Creating TIN dataset: ' + gradientTIN)
                arcpy.ddd.CreateTin(gradientTIN, inputSpatialReference, "spatialGradient Gradient masspoints", "DELAUNAY") 
                arcpy.AddMessage('TIN dataset created')
                logCursor.insertRow(('TIN dataset created', datetime.datetime.now()))

                spatialGradientSurface = 'spatialGradient' + suffix
                spatialGradientSurfaceTemp = spatialGradientSurface + '_temp' 
                arcpy.env.mask = inputSurveyArea 
                arcpy.TinRaster_3d(gradientTIN,  spatialGradientSurfaceTemp, "FLOAT", "LINEAR", "CELLSIZE " + str(inputCellSize), "1")
                
                temp = ExtractByMask(spatialGradientSurfaceTemp, inputSurveyArea)
                temp.save(spatialGradientSurface)
                
                arcpy.Delete_management(spatialGradientSurfaceTemp)
                arcpy.AddMessage('TIN surface raster created')

                arcpy.Delete_management(gradientTIN)
                arcpy.AddMessage('TIN deleted:' + gradientTIN)

            except arcpy.ExecuteError as e:
                arcpy.AddError('Tool Error Creating TIN dataset: ' + e.message)
                logCursor.insertRow(('tool error creating TIN dataset: ' + e.message, datetime.datetime.now()))
                return
            except Exception as e:
                arcpy.AddError('Error Creating TIN dataset: ' + e.message)
                logCursor.insertRow(('error creating TIN dataset: ' + e.message, datetime.datetime.now()))
                return
            arcpy.CheckInExtension("3D")
        else :
            arcpy.AddMessage('Unknown interpolation method: ' + inputSurfaceMethod)
            logCursor.insertRow(('Unknown Interpolation Method', datetime.datetime.now()))
            return

        arcpy.env.workspace = outputWorkspace 

        # Create contours of the raw gamma values 
        if arcpy.Exists(totalFieldSurface):
            Contour(totalFieldSurface, 'totalField' + suffix_w_raw_contour, inputRawContourInterval) 
            arcpy.AddMessage('Created total field contour lines')
            logCursor.insertRow(('Created total field contour lines', datetime.datetime.now()))
        else :
            arcpy.AddMessage('Could not find surface : ' + totalFieldSurface + ' in workspace: ' + arcpy.env.workspace)

        # Create contours of the gradient gamma values 
        if arcpy.Exists(spatialGradientSurface):
            Contour(spatialGradientSurface, 'spatialGradient' + suffix_w_gradient_contour, inputGradientContourInterval, inputGradientContourInterval / 2.0) 
            arcpy.AddMessage('Created spatial gradient contour lines')
            logCursor.insertRow(('Created spatial gradient contour lines', datetime.datetime.now()))
        else :
            arcpy.AddMessage('Could not find surface : ' + spatialGradientSurface + ' in workspace: ' + arcpy.env.workspace)

        # Delete the logCursor object
        del logCursor
        arcpy.CheckInExtension('Spatial') 

class DeleteIdenticalTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class)."""
        self.label = 'Delete Identical Points'
        self.description = '' 
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""
        param0 = arcpy.Parameter(displayName = 'Input Survey Point Feature Class', 
                                 name = 'inputSurveyFeatures', 
                                 datatype = 'DEFeatureClass', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param1 = arcpy.Parameter(displayName = 'Spatial Tolerance ([dist] [unit])', 
                                 name = 'tolerance', 
                                 datatype = 'String', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        # Default values 
        param1.value = '0.01 METERS'
        return [param0, param1] 

    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.SetProduct('arcinfo') == "NotLicensed": 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        # Set input variables 
        inputSurveyFeatures = parameters[0].valueAsText 
        arcpy.AddMessage('input Survey Features: {}'.format(inputSurveyFeatures))

        inputTolerance = parameters[1].value
        arcpy.AddMessage('Tolerance: {}'.format(inputTolerance))

        arcpy.management.DeleteIdentical(inputSurveyFeatures,"SHAPE", inputTolerance, 0)
        arcpy.AddMessage(arcpy.GetMessages())
        arcpy.AddMessage("Tool Complete")

class SurveyAreaTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class)."""
        self.label = 'Generate Survey Boundary'
        self.description = '' 
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""
        param0 = arcpy.Parameter(displayName = 'Input Survey Line or Point Feature Class', 
                                 name = 'inputSurveyFeatures', 
                                 datatype = 'DEFeatureClass', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param1 = arcpy.Parameter(displayName = 'Buffer Distance (meter)', 
                                 name = 'inputBufferDistance', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param2 =  arcpy.Parameter(displayName = 'Send Outputs to same geodatabase as input?', 
                                  name = 'useInputGeodatabase', 
                                  datatype = 'GPBoolean', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param3 = arcpy.Parameter(displayName = 'Existing Output File Geodatabase',
                                  name = 'outputWorkspace', 
                                  datatype = 'Workspace', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        # Default values 
        param1.value = 5.0
        param2.value = True
        return [param0, param1, param2, param3] 

    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.SetProduct('arcinfo') == "NotLicensed": 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        """Modify the values and properties of parameters before internal 
        validation is performed.  This method is called whenever a parameter 
        has been changed."""

        if (parameters[2].value == True):
            parameters[3].enabled = False

            if (parameters[0].altered and len(parameters[0].valueAsText) > 0) : 
                workspaceparts = os.path.split(parameters[0].valueAsText)
                parameters[3].value = workspaceparts[0]
            else :
                parameters[3].value = ''
        else :
            parameters[3].enabled = True

    def updateMessages(self, parameters): 
        """Modify the messages created by internal validation for each tool 
        parameter.  This method is called after internal validation."""
        
        if parameters[1].value < 3.0 :
            parameters[1].setErrorMessage('Buffer distance must be 3 meters or greater')

        try:
            arcpy.env.workspace = parameters[3].valueAsText
            if len(arcpy.env.workspace) > 0 and arcpy.Exists('surveyArea'):
                parameters[0].setWarningMessage('Buffer already defined. Datasets may be overwritten')
        except Exception as e:
            return

        try:
            inputFeatureClass = parameters[0].valueAsText
            if len(inputFeatureClass) > 0: 
                desc = arcpy.Describe(inputFeatureClass)

                if ((desc.shapeType != 'Point') and (desc.shapeType != 'Polyline')):
                    parameters[0].setErrorMessage('Incorrect input shape type, should be point or line feature class')

        except Exception as e:
            return

        return

    def execute(self, parameters, messages): 
        """The source code of the tool."""
        # Set input variables 
        inputSurveyFeatures = parameters[0].valueAsText 
        arcpy.AddMessage('input Survey Features: {}'.format(inputSurveyFeatures))

        inputBufferDistance = parameters[1].value
        arcpy.AddMessage('inputBufferDistance: {}'.format(str(inputBufferDistance)))

        outputWorkspace = parameters[3].valueAsText

        arcpy.AddMessage('output Geodatabase: {}'.format(outputWorkspace))

        # Set arcpy environment variables 
        arcpy.env.workspace = outputWorkspace 
        arcpy.env.overwriteOutput = True
        arcpy.AddMessage('Set environment variables') 

        # Create the logTable
        now = datetime.datetime.now()
        logTable = 'SurveyAreaLog_{}'.format(now.strftime('%Y%m%d%H%M%S'))
        arcpy.CreateTable_management(outputWorkspace, logTable)
        arcpy.management.AddField(logTable, 'Process', 'TEXT')
        arcpy.management.AddField(logTable, 'Completed', 'TEXT')

        # Create an insert cursor to write to the logTable
        logCursor = arcpy.da.InsertCursor(logTable,('Process', 'Completed'))

        # Write the input parameters to the logTable
        logCursor.insertRow(('Survey Features: {}'.format(inputSurveyFeatures), datetime.datetime.now()))
        logCursor.insertRow(('Buffer Distance: {}'.format(inputBufferDistance), datetime.datetime.now()))

        # Make convex hull 
        arcpy.management.MinimumBoundingGeometry(inputSurveyFeatures, 'convexHull', 'CONVEX_HULL') 
        arcpy.AddMessage('Made convex hull') 
        logCursor.insertRow(('Made convex hull', datetime.datetime.now()))

        # Create survey area by buffering convex hull 
        arcpy.analysis.Buffer('convexHull', 'convexHullArea', inputBufferDistance, 'FULL', '', 'ALL') 
        arcpy.AddMessage('Made survey area by buffering convex hull')
        if arcpy.Exists('convexHull'):
            arcpy.management.Delete('convexHull')
        logCursor.insertRow(('Made survey area by buffering convex hull', datetime.datetime.now()))

        arcpy.AddMessage('Creating Blanking Distance Coverage Polygon') 

        arcpy.analysis.Buffer(inputSurveyFeatures, 'surveyArea_temp1', inputBufferDistance, 'FULL', 'ROUND', 'ALL')

        arcpy.AddMessage('bufferred individual input features') 
        logCursor.insertRow(('bufferred individual input features', datetime.datetime.now()))

##        try:
##        
##            arcpy.management.EliminatePolygonPart('surveyArea_temp1', 'surveyArea_temp2', 'PERCENT', '', 99.0, 'CONTAINED_ONLY')
##
##            arcpy.AddMessage('Eliminated internal voids in search area') 
##            logCursor.insertRow(('Eliminated internal voids in search area', datetime.datetime.now()))
##        except Exception:
##            arcpy.AddMessage('Unable to eliminate internal voids in search area') 
##            logCursor.insertRow(('Unable to eliminate internal voids in search area', datetime.datetime.now()))

        tempfeature = None

        try:
            if arcpy.Exists('surveyArea_temp2'):
                tempfeature = 'surveyArea_temp2'
            else:
                tempfeature = 'surveyArea_temp1'

            arcpy.cartography.SimplifyPolygon(tempfeature, 'coverageArea', "POINT_REMOVE", (inputBufferDistance / 4.0), 0, 'RESOLVE_ERRORS', 'NO_KEEP')

            arcpy.AddMessage('Simplified blanking distance coverage area') 
            logCursor.insertRow(('Simplified bufferred survey area', datetime.datetime.now()))
        except Exception:
            arcpy.AddMessage('Unable to simplify survey area') 
            logCursor.insertRow(('Unable to simplify survey area', datetime.datetime.now()))

            arcpy.management.Rename(tempfeature, 'coverageArea')

        if arcpy.Exists('coverageArea' + '_Pnt'):
            arcpy.management.Delete('coverageArea' + '_Pnt')

        if arcpy.Exists('surveyArea_temp1'):
            arcpy.management.Delete('surveyArea_temp1')
        
        if arcpy.Exists('surveyArea_temp2'):
            arcpy.management.Delete('surveyArea_temp2')

        logCursor.insertRow(('Made Blanking Distance Polygon', datetime.datetime.now()))

        # Delete the logCursor object
        del logCursor

class SmoothRawNavigationTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class)."""
        self.label = 'Smooth Raw Navigation'
        self.description = '' 
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""
        param0 = arcpy.Parameter(displayName = 'Input Raw File', 
                                 name = 'inputRawFile', 
                                 datatype = 'Table', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param1 = arcpy.Parameter(displayName = 'Rolling Stats Window Size', 
                                 name = 'window', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param2 = arcpy.Parameter(displayName = 'X field', 
                                 name = 'xfield', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param3 = arcpy.Parameter(displayName = 'Y field', 
                                 name = 'yfield', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param4 = arcpy.Parameter(displayName = 'Line field', 
                                 name = 'inputLinefield', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param5 = arcpy.Parameter(displayName = 'NavSmth Temp Directory', 
                                 name = 'navSmthPathway', 
                                 datatype = 'Workspace', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param6 = arcpy.Parameter(displayName = 'Output Pathway', 
                                 name = 'outputPathway', 
                                 datatype = 'Workspace', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param7 = arcpy.Parameter(displayName = 'Output File Name (append _smth)', 
                                 name = 'outputFileName', 
                                 datatype = 'String', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 
        # Default values 
        param1.value = 5
        param2.value = "Field1"
        param3.value = "Field2"
        param6.value = "outfile"

        param2.parameterDependencies = [param0.name]
        param3.parameterDependencies = [param0.name]
        param4.parameterDependencies = [param0.name]
        return [param0, param1, param2, param3, param4, param5, param6, param7] 


    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.SetProduct('arcinfo') == "NotLicensed": 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        # Set input variables 
        inputRawFile = parameters[0].valueAsText 
        arcpy.AddMessage('Raw File: {}'.format(inputRawFile))

        window=(int(parameters[1].value))
        arcpy.AddMessage('Rolling Stats Window Size: {} points'.format(window))

        inputXfield=parameters[2].valueAsText

        inputYfield=parameters[3].valueAsText

        inputLinefield=parameters[4].valueAsText

        navSmthPathway=parameters[5].valueAsText

        outputWorkspace=parameters[6].valueAsText

        outputFileName=parameters[7].valueAsText 

        inputTable=pd.read_csv(r'{}'.format(inputRawFile))
        df=pd.DataFrame(inputTable)
        arcpy.AddMessage("Read raw file")

        ##iterate a list of unique line names
        lineNameList=df['{}'.format(inputLinefield)].unique()
        arcpy.AddMessage(lineNameList)
        arcpy.AddMessage("Total Number of Lines: {}".format(len(lineNameList)))

        ## define a function to perform a rolling stats computation on single line of data
        def rollingStatsperLine(x): 
            arcpy.AddMessage("Line Name: {}".format(lineNameList[x]))
            mask=df['{}'.format(inputLinefield)]==lineNameList[x]
            dfx=df[mask].copy()
            dfx['{}_smth'.format(inputXfield)]=dfx['{}'.format(inputXfield)].rolling(window,center=True).mean()
            dfx['{}_smth'.format(inputYfield)]=dfx['{}'.format(inputYfield)].rolling(window,center=True).mean()
            dfx.to_csv(r"{}\{}{}.txt".format(navSmthPathway,outputFileName,x),float_format='%.2f',index=False)
            return

        ## call the function and run as iterations of the line name list
        x=0
        while x <= len(lineNameList):
            rollingStatsperLine(x)
            x +=1
            if x == len(lineNameList):
                break

        ## read directory of CSV files generated for each df split by line, merge into a single df
        path =(r"{}".format(navSmthPathway))
        all_files = glob.glob(os.path.join(path,"{}*.txt".format(outputFileName)))
        df_from_each_file = (pd.read_csv(f,sep=',')for f in all_files)
        df_merged = pd.concat(df_from_each_file,ignore_index=True)
        df_final=df_merged.dropna()
        arcpy.AddMessage("Merged Resutls")

        ## export merged df as full csv file
        os.chdir(r"{}".format(outputWorkspace))
        df_final.to_csv("{}_navsmth.txt".format(outputFileName),index=False)

        for f in all_files:
            os.remove(f)

        arcpy.AddMessage("completed line smoothing")

class SmoothRawAltitudeTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class)."""
        self.label = 'Smooth Raw Altitude'
        self.description = '' 
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""
        param0 = arcpy.Parameter(displayName = 'Input Raw File', 
                                 name = 'inputRawFile', 
                                 datatype = 'Table', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param1 = arcpy.Parameter(displayName = 'Altitude Max Threshold', 
                                 name = 'altMax', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param2 = arcpy.Parameter(displayName = 'Rolling Stats Window Size', 
                                 name = 'window', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param3 = arcpy.Parameter(displayName = 'Altitude field', 
                                 name = 'inputAltfield', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')

        param4 = arcpy.Parameter(displayName = 'Line field', 
                                 name = 'inputLinefield', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param5 = arcpy.Parameter(displayName = 'Time field', 
                                 name = 'inputTimefield', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input')                         

        param6 = arcpy.Parameter(displayName = 'AltSmth Temp Directory', 
                                 name = 'altSmthPathway', 
                                 datatype = 'Workspace', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param7 = arcpy.Parameter(displayName = 'Output Pathway', 
                                 name = 'outputPathway', 
                                 datatype = 'Workspace', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param8 = arcpy.Parameter(displayName = 'Output File Name (tool will append _Altsmth)', 
                                 name = 'outputFileName', 
                                 datatype = 'String', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 
        
        # Default values 
        param1.value = 20
        param2.value = 5
        param3.value = "Field2"
        param4.value = "Field3"
        param5.value = "Field4"
        param6.value = "outfile"

        param3.parameterDependencies = [param0.name]
        param4.parameterDependencies = [param0.name]
        param5.parameterDependencies = [param0.name]
        return [param0, param1, param2, param3, param4, param5, param6, param7,param8] 


    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.SetProduct('arcinfo') == "NotLicensed": 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages):
        # Set input variables 
        inputRawFile = parameters[0].valueAsText 
        arcpy.AddMessage('Raw File: {}'.format(inputRawFile))

        altMax=(int(parameters[1].value))
        arcpy.AddMessage('Altitude Maximum Threshold: {}'.format(altMax))

        window=(int(parameters[2].value))
        arcpy.AddMessage('Rolling Stats Window Size: {} points'.format(window))

        inputAltfield=parameters[3].valueAsText

        inputLinefield=parameters[4].valueAsText

        inputTimefield=parameters[5].valueAsText

        altSmthPathway=parameters[6].valueAsText

        outputWorkspace=parameters[7].valueAsText

        outputFileName=parameters[8].valueAsText 

        inputTable=pd.read_csv(r'{}'.format(inputRawFile))
        df=pd.DataFrame(inputTable)
        arcpy.AddMessage("Read raw file")

        # plot raw altitude; use this to confirm or adjust max alt threshold
        plt.style.use('classic')
        x=df['{}'.format(inputTimefield)]
        y=df['{}'.format(inputAltfield)]
        fig,ax=plt.subplots()
        ax.plot(x,y, linewidth=1.0)
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(True)
        plt.savefig(r"{}\{}.png".format(altSmthPathway,outputFileName))

        ##iterate a list of unique line names
        lineNameList=df['{}'.format(inputLinefield)].unique()
        arcpy.AddMessage(lineNameList)
        arcpy.AddMessage("Total Number of Lines: {}".format(len(lineNameList)))

        df['{}_dspk'.format(inputAltfield)]=df['{}'.format(inputAltfield)].where(df['{}'.format(inputAltfield)]< altMax)

        ## define a function to perform a rolling stats computation on single line of data
        def rollingStatsperLine(x): 
            arcpy.AddMessage("Line Name: {}".format(lineNameList[x]))
            mask=df['{}'.format(inputLinefield)]==lineNameList[x]
            dfx=df[mask].copy()
            dfx['{}_intp'.format(inputAltfield)]=dfx['{}_dspk'.format(inputAltfield)].interpolate(method='linear')
            dfx['{}_smth'.format(inputAltfield)]=dfx['{}_intp'.format(inputAltfield)].rolling(window,center=True).mean()
            dfx.to_csv(r"{}\{}ln{}.txt".format(altSmthPathway,outputFileName,x),float_format='%.2f',index=False)
        ##plot raw channel
            plt.style.use('classic')
            a=dfx['{}'.format(inputTimefield)]
            b=dfx['{}'.format(inputAltfield)]
            fig,ax=plt.subplots()
            ax.plot(a,b, linewidth=1.0)
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(True)
            plt.savefig(r"{}\1_raw_{}_ln{}.png".format(altSmthPathway,outputFileName,x))
        ##plot despike channel result 
            plt.style.use('classic')
            c=dfx['{}'.format(inputTimefield)]
            d=dfx['{}_dspk'.format(inputAltfield)]
            fig,ax=plt.subplots()
            ax.plot(c,d, linewidth=1.0)
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(True)
            plt.savefig(r"{}\2_dspk_{}_ln{}.png".format(altSmthPathway,outputFileName,x))
        ##plot smoothed channel result
            plt.style.use('classic')
            e=dfx['{}'.format(inputTimefield)]
            f=dfx['{}_smth'.format(inputAltfield)]
            fig,ax=plt.subplots()
            ax.plot(e,f, linewidth=1.0)
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(True)
            plt.savefig(r"{}\3_smth_{}_ln{}.png".format(altSmthPathway,outputFileName,x))
            return


        ## call the function and run as iterations of the line name list
        x=0
        while x <= len(lineNameList):
            rollingStatsperLine(x)
            x +=1
            if x == len(lineNameList):
                break

        ## read directory of CSV files generated for each df split by line, merge into a single df
        path =(r"{}".format(altSmthPathway))
        all_files = glob.glob(os.path.join(path,"{}*.txt".format(outputFileName)))
        df_from_each_file = (pd.read_csv(f,sep=',')for f in all_files)
        df_merged = pd.concat(df_from_each_file,ignore_index=True)
        df_final=df_merged.dropna()
        arcpy.AddMessage("Merged Resutls")

        ## export merged df as full csv file
        os.chdir(r"{}".format(outputWorkspace))
        df_final.to_csv("{}_Altsmth.txt".format(outputFileName),index=False)

        for f in all_files:
            os.remove(f)

        arcpy.AddMessage("completed altitude smoothing")